/*
 * (c) Copyright 2001
 * Peter Blum, Peter.Blum@readme1st.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>

#include "m68000.h"
#include "device.h"
#include "conf.h"


/* Address decoder table */
static int    devNum;			/* Current number of devices */
static Device devTable[MAXDEV];		/* Device table */

static int    decResetCyc;              /* Reset Cycle counter */
static u_char intPending;               /* Pending interrupts */


/*
 *  Call cycle routine of each device
 */
void decCycle(int sig)
{
    Device  *dev;
    int  i;

    for (i = 0; i < devNum; i++) {
        dev = &devTable[i];
        if (dev->devCycle != NULL)  dev->devCycle(dev);
    }
}


/*
 *  R e s e t
 */
int decoderReset(void)
{
    int   i;

    if (!quiet)  printf("Reseting all devices ... ");

    /* Call reset function of all devices */
    for (i = 0; i < devNum; i++)  {
        devTable[i].devReset(&devTable[i]);
    }

    if (!quiet)  printf("\n");

    return 0;
}


/*
 * Search the device table for a address match
 */
static Device *decFindDev(u_long addr, u_long size)
{
    Device *dev = &devTable[0];
    Device *end = &devTable[devNum];

    for (; dev < end; dev++)  {
        if ((addr & dev->mask) == dev->base)  {
            return dev;
        }
    }

    return NULL;
}


/*
 * R e a d   f r o m   m e m o r y
 */
int memRead(u_long addr, u_long size, u_long *result)
{
    Device  *dev;
    u_long  temp, offset;

    if (decResetCyc)  {
        addr += conf.rstBase;
        decResetCyc -= 2;
    }

    if ((dev = decFindDev(addr, size)) == NULL)
        return BUS_ERROR;

    /* Calculate offset within device address space */
    offset = addr - dev->base;

    /* Read from device */
    if (size == LONG)  {
        temp = dev->devRead(dev, offset+0);
        temp = (temp << 8) | dev->devRead(dev, offset+1);
        temp = (temp << 8) | dev->devRead(dev, offset+2);
        temp = (temp << 8) | dev->devRead(dev, offset+3);
    }
    else if (size == WORD)  {
        temp = dev->devRead(dev, offset+0);
        temp = (temp << 8) | dev->devRead(dev, offset+1);
    }
    else  {
        temp = dev->devRead(dev, offset+0);
    }
    *result = temp;

    return SUCCESS;
}


/*
 *  W r i t e   t o   m e m o r y
 */
int memWrite(u_long addr, u_long size, u_long data)
{
    Device  *dev;
    u_long  offset;

    if ((dev = decFindDev(addr, size)) == NULL)
        return BUS_ERROR;

    /* Calculate offset within device address space */
    offset = addr - dev->base;

    if (size == LONG)  {
        dev->devWrite(dev, offset+0, (u_char)((data >> 24) & 0xFF));
        dev->devWrite(dev, offset+1, (u_char)((data >> 16) & 0xFF));
        dev->devWrite(dev, offset+2, (u_char)((data >> 8)  & 0xFF));
        dev->devWrite(dev, offset+3, (u_char)((data >> 0)  & 0xFF));
    }
    else if (size == WORD)  {
        dev->devWrite(dev, offset+0, (u_char)((data >> 8) & 0xFF));
        dev->devWrite(dev, offset+1, (u_char)((data >> 0) & 0xFF));
    }
    else  {
        dev->devWrite(dev, offset+0, (u_char)((data >> 0) & 0xFF));
    }

    return SUCCESS;
}


/*
 *  I n t e r r u p t   A c k n o w l e d g e
 */
u_char intAcknowledge(int level)
{
    Device *dev;
    int    i;
    u_char vector = 0x00;
	

    if (!intPending)
        /* Return spurious interrupt vector */
        return 24;

    for (i = 0; i < devNum; i++)  {
        dev = &devTable[i];
        if (level == dev->intLevel)  {
            vector = dev->devIAck(dev, level);
        }
    }

    return vector;
}


/*
 *  S e t   I n t e r r u p t   R e q u e s t
 */
void intSetIRQ(Device *dev)
{
    int  i;

    intPending |= (1 << dev->intLevel);

    for (i = 7; i > 0; i--)  {
        if (intPending & (1 << i))  break;
    }

    cpuSetIPL(i);
}


/*
 *  C l e a r   I n t e r r u p t   R e q u e s t
 */
void intClearIRQ(Device *dev)
{
    int  i;

    intPending &= ~(1 << dev->intLevel);

    for (i = 7; i > 0; i--)  {
        if (intPending & (1 << i))  break;
    }
    cpuSetIPL(i);
}


/*
 *  I n i t i a l i z e
 */
int decoderInit(void)
{
    int   i;
    int   (*initFunc)(Device *dev, u_long base, u_long size, void *arg);
    Dev   *dp;

    for (i = 0; i < conf.numDev; i++)  {
        dp = &conf.dev[i];
        if (!strcasecmp("ram", dp->type))
            initFunc = initRAM;
        else if (!strcasecmp("eprom", dp->type))
            initFunc = initEPROM;
        else if (!strcasecmp("m68681", dp->type))
            initFunc = initM68681;
        else if (!strcasecmp("m68230", dp->type))
            initFunc = initM68230;
        else if (!strcasecmp("ata", dp->type))
            initFunc = initATA;
        else if (!strcasecmp("mk48t02", dp->type))
            initFunc = initMK48T02;
        else {
            fprintf(stderr, "ERROR: Unknown device '%s'\n", dp->type);
            return(-1);
        }
        if (initFunc(&devTable[i], dp->base, dp->size, dp->devarg) != 0)  {
            fprintf(stderr, "ERROR: Unable to initialize %s\n", dp->type);
            return(-1);
        }
        devTable[i].intLevel = dp->intlvl;
        if (!quiet)
            printf("%-10s at %08lx %08lx\n", dp->type, dp->base, dp->size);
    }

    devNum = i;

    /* No pending interrupts */
    intPending = 0;

    /* Set reset cycle flag */
    decResetCyc = 4;

    return 0;
}



