/*
 * (c) Copyright 2001
 * Peter Blum, Peter.Blum@readme1st.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>

#include "device.h"
#include "conf.h"


static void epromLoad(char *fName, u_char *eprom, u_long base, u_long size);



/*
 *  R e s e t
 */
int epromReset(Device *dev)
{
    if (!quiet)  printf(" EPROM");

    return 0;
}


/*
 *  R e a d   o n e   b y t e
 */
static u_long epromRead(Device *dev, u_long offset)
{
    u_char *eprom = (u_char *) dev->priv;

    return (u_long) eprom[offset];
}


/*
 *  W r i t e   o n e   b y t e
 */
static void epromWrite(Device *dev, u_long offset, u_char data)
{
    u_char *eprom = (u_char *) dev->priv;

    eprom[offset] = data;
}


/*
 *  I n i t i a l i z e
 */
int initEPROM(Device *dev, u_long base, u_long size, void *arg)
{
    u_char  *eprom;

    if ((eprom = malloc(size)) == NULL) {
        fprintf(stderr, "ERROR: Malloc of %ld bytes failed\n", size);
        exit(errno);
    }

    if (arg != NULL)
        epromLoad((char *) arg, eprom, base, size);

    dev->base = (u_long) base;
    dev->mask = ~(size-1);
    dev->priv = (void *) eprom;
    dev->devReset = epromReset;
    dev->devRead = epromRead;
    dev->devWrite = epromWrite;
    dev->devIAck  = NULL;
    dev->devCycle = NULL;

    return 0;
}


/*
 *  Read line from file.
 */
static char *epromReadLine(char *bp, int sz, FILE *f)
{
    int  c, i;

    for (i = 0; i < sz; i++)  {
        c = getc(f);
        if ((c == '\r') || (c == '\n')) {
            bp[i] = '\0';
            break;
        }
        bp[i] = (char) c;
    }

    return bp;
}


/*
 *  Load memory with contents of S-Record file
 */
static void epromLoadSRec(char *fName, u_char *eprom, u_long base, u_long size)
{
    FILE   *fp;
    char   buffer[128];
    char   *bptr, type;
    int    i, line = 0;	
    u_long address, nbyte, data;

    if ((fp = fopen(fName, "r")) == NULL)  {
        fprintf(stderr, "eprom: Can't open %s\n", fName);
        return;
    }

    while (!feof(fp))  {
        line++;
        if ((bptr = epromReadLine(buffer, 128, fp)) == NULL)  {
            fprintf(stderr, "Unexpected end of file - %s\n", strerror(errno));
            return;
        }

        /* Get Type */
        if (sscanf(bptr, "S%c", &type) != 1)  {
            /* Ignore lines that don't start with S */
            break;
        }
        bptr += 2;

        /* Get Length */
        if (sscanf(bptr, "%02lx", &nbyte) != 1)  {
            fprintf(stderr, "Line %d: Bad character in length\n", line);
            return;
        }
        bptr += 2;

        /* Get address field */
        switch (type)  {
            case '0' :
                /* Ignore start record */
                continue;

            case '1' :
                if (sscanf(bptr,"%04lx", &address) != 1)  {
                    fprintf(stderr, "Line %d: Bad address\n", line);
                    return;
                }
                bptr += 4;
                nbyte -= 2;
                break;

            case '2' :
                if (sscanf(bptr,"%06lx", &address) != 1)  {
                    fprintf(stderr, "Line %d: Bad address\n", line);
                    return;
                }
                bptr += 6;
                nbyte -= 3;
                break;

            case '3' :
                if (sscanf(bptr,"%08lx", &address) != 1)  {
                    fprintf(stderr, "Line %d: Bad address\n", line);
                    return;
                }
                bptr += 8;
                nbyte -= 4;
                break;

            case '7' :
            case '8' :
            case '9' :
                /* Ignore all termination records */
                continue;

            default :
                fprintf(stderr, "Line %d: Bad type '%c'\n", line, type);
                return;
        }

        /* Get all data bytes without checksum */
        for (i = 0; i < nbyte-1; i++)  {
            if (sscanf(bptr, "%02lx", &data) != 1)  {
                fprintf(stderr, "ERROR: Line %d: Bad character in data\n", line);
                return;
            }
            bptr += 2;
            if ((address >= base) || (address < (base + size)))  {
                eprom[address - base] = (u_char) data;
                address++;
            }
            else  {
                fprintf(stderr, "ERROR: Address %08lx out of range", address);
                return;
            }
        }

        /* Ignore checksum */
        if (sscanf(bptr, "%02lx", &data) != 1)  {
            fprintf(stderr, "ERROR: Line %d: Bad character in checksum\n", line);
            return;
        }
    }

    fclose(fp);
}


/*
 *  Load memory with contents of Binary file
 */
static void epromLoadBin(char *fName, u_char *eprom, u_long base, u_long size)
{
    FILE   *fp;
    char   buffer[128];
    int    i;
    u_long address = 0;
    u_long nbyte;

    if ((fp = fopen(fName, "r")) == NULL)  {
        fprintf(stderr, "ld: Can't open %s\n", fName);
        return;
    }

    while (!feof(fp))  {
        nbyte = fread(buffer, 128, sizeof(char), fp);
        for (i = 0; i < nbyte; i++)  {
            if (address >= size)  {
                fprintf(stderr, "ERROR: Reached end of EPROM\n");
                return;
            }
            eprom[address++] = (u_char) buffer[i];
        }
    }

    fclose(fp);
}

/*
 *  Load memory with contents of Binary file
 */
static void epromLoad(char *fName, u_char *eprom, u_long base, u_long size)
{
    FILE   *fp;
    int    c;

    if ((fp = fopen(fName, "r")) == NULL)  {
        fprintf(stderr, "eprom: Can't open %s\n", fName);
        return;
    }

    c = getc(fp);
    fclose(fp);

    if (c == (int) 'S')
        epromLoadSRec(fName, eprom, base, size);
    else
        epromLoadBin(fName, eprom, base, size);
}
