/*
 * (c) Copyright 2001
 * Peter Blum, Peter.Blum@readme1st.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>

#include "m68000.h"
#include "decoder.h"
#include "opcodes.h"
#include "conf.h"


int   trace;                    /* Enable disassembling */
long  D[8], A[9], PC, SR;       /* Processor Register */
int   cpuState;                 /* CPU state */
long  inst;                     /* Current instruction */

static int   intLevel;          /* Current interrupt level pending */
static int   cpuCounter;

static int exception(int vector, int class, long addr, int r_w);


/*
 *  R e s e t   t h e   M 6 8 0 0 0
 */
int cpuReset(void)
{
    SR = 0x2700;
    cpuState = ST_RESET;

    if (memRead(0x000000, LONG, &A[8]) || memRead(0x000004, LONG, &PC))  {
        fprintf(stderr, "ERROR: Bus Error during Reset vector read\n");
        cpuState = ST_HALT;
        return 0;
    }
    cpuState = ST_RUNNING;
 
    return 0;
}


/*
 *  I n i t i a l i z   t h e   M 6 8 0 0 0
 */
int cpuInit(void)
{
    int   i;

    for (i = 0; i <= 7; i++)
        A[i] = D[i] = 0x00000000;
    A[8] = 0x00000000;
    PC = 0x00000000;
    SR = 0x2700;

    intLevel = 0;                      /* No pending interrupt */
    cpuCounter = conf.numInstCycle;    /* Preset polling counter */

    return 0;
}


/*
 *  P u t   d a t a   t o   m e m o r y
 * 
 *  This function also performs the exception processing.
 */
int memPut(long data, long addr, long size)
{
    if ((addr & 0x000001) && (size != BYTE))  {
        exception(3, 0, addr, WRITE);
        return ADDRESS_ERROR;
    }

    if (memWrite(addr, size, data))  {
        exception(2, 0, addr, WRITE);
        return BUS_ERROR;
    }

    return SUCCESS;
}


/*
 * G e t   d a t a   f r o m   m e m o r y
 *
 * This function also performs the exception handling
 */
int memGet(long addr, long size, long *result)
{
    u_long  temp;

    if ((addr & 0x000001) && (size != BYTE))  {
        exception(3, 0, addr, READ);
        return ADDRESS_ERROR;
    }

    if (memRead(addr, size, &temp))  {
        exception(2, 0, addr, READ);
        return BUS_ERROR;
    }

    *result = (long)temp;

    return SUCCESS;
}


/*
 * G e t   d a t a   f r o m   c o d e   a r e a
 */
int memRequest(long *addrPtr, long size, long *result)
{
    u_long  addr = *addrPtr;
    u_long  temp;

    if ((addr & 0x000001) && (size != BYTE))  {
        exception(3, 0, addr, READ);
        return ADDRESS_ERROR;
    }

    if (memRead(addr, (size == BYTE) ? WORD : size, &temp))  {
        exception(2, 0, addr, READ);
        return BUS_ERROR;
    }

    *result = temp;

    if (size == LONG)  {
        addr += 4;
    }
    else if (size == WORD) {
        addr += 2;
    }
    else  {
        *result &= 0x00ff;
        addr += 2;
    }
    *addrPtr = addr;

    return SUCCESS;
}



/*
 *  E x c e p t i o n   h a n d l i n g
 */
static int exception(int vector, int class, long addr, int r_w)
{
    int   iword;
    long  tmpSR;

    tmpSR = SR;

    SR = SR | sbit;     /* force processor into supervisor state */
    SR = SR & ~tbit;    /* turn off trace mode */

    if ( (class == 1) || (class == 2))  {
        A[8] -= 4; memPut(PC, A[8], LONG);
        A[8] -= 2; memPut(tmpSR, A[8], WORD);
    }
    else {
        iword = (r_w == READ) ? 0x0036 : 0x0006;
        A[8] -= 4; memPut(PC, A[8], LONG);
        A[8] -= 2; memPut(tmpSR, A[8], WORD);
        A[8] -= 2; memPut(inst, A[8], WORD);
        A[8] -= 4; memPut(addr, A[8], LONG);
        A[8] -= 2; memPut(iword, A[8], WORD);
    }

    memGet(vector << 2, LONG, &PC);

    return SUCCESS;
}



/*
 *  E x e c u t e s   a   s i n g l e   i n s t r u c t i o n
 */
void cpuExec(void)
{
    int  vector;
    int  idx, funcRet;
    Inst *instPtr, *endPtr;
    int  ret;

    if (--cpuCounter == 0)  {
        /* Call decoders cycle function */
        decCycle();
        cpuCounter = conf.numInstCycle;
    }

    if (cpuState == ST_HALT)  {
        if (trace)  printf("CPU Halted\n");
        return;
    }

    if (intLevel && (intLevel > ((SR & 0x0700) >> 8)))  {
        vector = intAcknowledge(intLevel);
        if (trace)  printf("Interrupt vector %02x\n", vector);
        exception(vector, 1, 0, READ);
        SR = ((SR & ~0x0700) | (intLevel << 8));
        cpuState = ST_RUNNING;
        return;
    }

    if (cpuState == ST_STOP)  {
        if (trace)  printf("CPU Stopped\n");
        return;
    }

    /* Get opcode from memory and execute */
    if ((ret = memGet(PC, WORD, &inst)) != SUCCESS)  {
        /* Bus or Address error */
        return;
    }

    if ((inst & 0xF000) == 0xA000)  {
        /* Unimplemented instruction */
        if (trace)  printf("%08lx: Line 1010\n", PC);
        exception(10, 1, 0, READ);
    }
    else if ((inst & 0xF000) == 0xF000)  {
        /* Unimplemented instruction */
        if (trace)  printf("%08lx: Line 1111 Trap\n", PC);
        exception(11, 1, 0, READ);
    }
    else  {
        /* Search function for this instruction */
        idx = (inst & 0xF000) >> 12;
        instPtr = &instTable[instOffset[idx]];
        endPtr  = &instTable[instOffset[idx+1] - 1];

        for (; instPtr <= endPtr; instPtr++)
            if ((inst & ~instPtr->mask) == instPtr->val)  break;
        if (instPtr > endPtr)  {
            printf("Opcode %04x not found\n", (u_short) inst);
            exit(0);
        }

        if (trace) printf("%08lx: %04lx %s", PC, inst, instPtr->disasm);

        /* Execute instruction */
        PC += 2;
        funcRet = instPtr->func();

        if (trace)  printf("\n");

        switch (funcRet)  {
            case SUCCESS  :
            case SUCCESS_PRIV  :
                break;

            case BAD_INST :
                if (trace)  printf("Illegal instruction at %08lx\n", PC - 2);
	        exception(4, 1, 0, READ);
                break;

            case NO_PRIVILEGE :
                if (trace)  printf("Privilege Violation at %08lx\n", PC - 2);
                exception(8, 1, 0, READ);
                memGet(0x20, LONG, &PC);
                break;

            case CHK_EXCEPTION :
                if (trace)  printf("CHK Instruction at %08lx\n", PC - 2);
                exception(6, 2, 0, READ);
                break;

            case ILLEGAL_TRAP :
                if (trace)  printf("ILLEGAL Instruction at %08lx\n", PC - 2);
                exception(4, 1, 0, READ);
                break;

            case STOP_TRAP :
                break;

            case TRAP_TRAP :
                if (trace)  printf("TRAP Instruction at %08lx\n", PC - 2);
                exception(32 + (inst & 0x000F), 2, 0, READ);
                break;

            case TRAPV_TRAP :
                if (trace)  printf("TRAPV Instruction at %08lx\n", PC - 2);
                exception(7, 2, 0, READ);
                break;

            case DIV_BY_ZERO :
                if (trace)  printf("Zero Divide at %08lx\n", PC - 2);
                exception(5, 2, 0, READ);
                break;
        }

        if ((funcRet != SUCCESS_PRIV) && (SR & tbit))  {
            if (trace)  printf("Trace Trap at %08lx\n", PC - 2);
            exception(9, 1, 0, READ);
        }
    }
}


/*
 *  R e q u e s t   i n t e r r u p t
 */
void cpuSetIPL(int level)
{
	intLevel = level;
}

