/*
 * (c) Copyright 2001
 * Peter Blum, Peter.Blum@readme1st.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>

#include "m68000.h"
#include "decoder.h"
#include "conf.h"
#include "ui.h"
#include "version.h"


char   *prgName = "s68";
char   *confFile = S68CONF;     /* Default conf file name */
Conf   conf;                    /* S68 configraion from conf file */
int    quiet = 0;               /* Quiet mode */



/*
 *  P r i n t   U s a g e
 */
int _usage(int e, char *s)
{
    if (s != NULL)
        fprintf(stderr, "%s: %s\n", prgName, s);
    fprintf(stderr, "M68000 System Simulator\n");
    fprintf(stderr, "Usage: %s [-h] [-V] [<cfg-file>]\n", prgName);
    fprintf(stderr, "  -h           Display this message\n");
    fprintf(stderr, "  -V           Print program version\n");

    return e;
}


/*
 *  R e a d   C o n f i g u r a t i o n
 */
int readConf(char *sFile)
{
    FILE  *sf;
    char  line[1024];
    char  *p;
    char  *pname, *pvalue;
    Dev   *dp;
    Cmd   *cp;

    if ((sf = fopen(sFile, "r")) == NULL)  return(-1);

    conf.numDev = 0;
    conf.numCmd = 0;
    conf.rstBase = 0x000000;
    conf.numInstCycle = 1000;

    while ((p = fgets(line, sizeof(line), sf)) != NULL)  {
        pname = strtok(p, " \t\n");
        if ((pname == NULL) || (pname[0] == '\0') || (pname[0] == '#'))
            continue;

        if (!strcasecmp(pname, "COMMAND"))  {
            /* Get pointer to next command pointer */
            cp = &conf.cmd[conf.numCmd++];

            /* Get command */
            pvalue = strtok(NULL, "\n");
            if ((*cp = malloc(strlen(pvalue)+1)) == NULL)  {
                fprintf(stderr, "ERROR: Out of memory\n");
                exit(1);
            }
            strcpy(*cp, pvalue);
        }
        else if (!strcasecmp(pname, "DEVICE"))  {
            /* Get pointer to next device structure */
            dp = &conf.dev[conf.numDev++];

	    /* Get device type */
            pvalue = strtok(NULL, " \t\n");
            if ((dp->type = malloc(strlen(pvalue)+1)) == NULL)  {
                fprintf(stderr, "ERROR: Out of memory\n");
                exit(1);
            }
            strcpy(dp->type, pvalue);
 
            /* Get base address */
            pvalue = strtok(NULL, " \t\n");
            sscanf(pvalue, "%lx", &dp->base);

            /* Get size */
            pvalue = strtok(NULL, " \t\n");
            sscanf(pvalue, "%lx", &dp->size);

            /* Get interrupt level */
            pvalue = strtok(NULL, " \t\n");
            if (pvalue == NULL)  {
                dp->intlvl = 0;
            }
            else  {
                if (sscanf(pvalue, "%d", &dp->intlvl) != 1)  {
                    fprintf(stderr, "ERROR: Illegal interrupt level");
                    exit(1);
                }
            }

            /* Get device arguments */
            pvalue = strtok(NULL, " \t\n");
            if (pvalue == NULL)  {
                dp->devarg = NULL;
            }
            else  {
                if ((dp->devarg = malloc(strlen(pvalue)+1)) == NULL)  {
                    fprintf(stderr, "ERROR: Out of memory\n");
                    exit(1);
                }
                strcpy(dp->devarg, pvalue);
            }
        }
        else if (!strcasecmp(pname, "CYCLE"))  {
            /* Get number of Inst. per Polling cycle */
            pvalue = strtok(NULL, " \t\n");
            sscanf(pvalue, "%d", &conf.numInstCycle);
        }
        else if (!strcasecmp(pname, "RSTBASE"))  {
            /* Get base address */
            pvalue = strtok(NULL, " \t\n");
            sscanf(pvalue, "%lx", &conf.rstBase);
        }
        else {
            fprintf(stderr, "WARNING: '%s' unknown keyword\n", pname);
        }
    }

    return 0;
}


/*
 *  M a i n
 */
int main(int argc, char *argv[])
{
    int   n;

    /* Get command line parameters */
    for (n = 1; n < argc; n++)  {
        if (argv[n][0] == '-')  {
            /* Command line option */
            switch (argv[n][1])  {
              case 'V' :
                /* Display version */
                fprintf(stderr, "%s %d.%d.%d [%s]\n", prgName, S68VER, UIVER, SUBVER, __DATE__);
                exit(0);

              case 'q' :
                  quiet = 1;
                  break;

              case 'h' :
                  exit(_usage(0, NULL));

              default :
                  exit(_usage(1, "Bad command line option\n"));
            }
        }
        else  {
            confFile = argv[n];
        }
    }

    if (!quiet)  {
        printf("M68000 System Simulator  ");
        printf("Version %d.%d.%d [%s]\n", S68VER, UIVER, SUBVER, __DATE__);
    }

    if (readConf(confFile)  < 0)  {
        fprintf(stderr, "WARNING: No configuration file\n");
    }

    decoderInit();      /* Initialize the Decoder and all I/O Devices */
    cpuInit();          /* Initialize the CPU */

    decoderReset();     /* Reset Decoder and all I/O Devices */
    cpuReset();         /* Reset CPU. Read reset vector and stack pointer */

    uiStart();          /* Start user interface */

    return 0;
}

