/*
 * (c) Copyright 2001
 * Peter Blum, Peter.Blum@readme1st.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <termios.h>
#include <fcntl.h>
#include <string.h>

#include "device.h"
#include "decoder.h"
#include "conf.h"



#define  REGSIZE   0x1000

#define  CNTL   0x0ff1          /* Control register */
#define  SEC    0x0ff3          /* seconds register */
#define  MIN    0x0ff5          /* minutes register */
#define  HOUR   0x0ff7          /* hours register */
#define  DW     0x0ff9          /* day of week register */
#define  DAY    0x0ffb          /* day register */
#define  MON    0x0ffd          /* month register */
#define  YEAR   0x0fff          /* year register */


typedef struct  {
    u_char  sram[0x7f8];
    u_char  cntl;
    u_char  sec;
    u_char  min;
    u_char  hour;
    u_char  dw;
    u_char  day;
    u_char  mon;
    u_char  year;
} MK48T02;


/*
 * Convert binary to bcd
 */
static u_char bin2bcd(int bin)
{
    return (u_char)(((bin / 10) << 4) | (bin % 10));
}

/*
 * Update RTC register from system clock
 */
static void update(MK48T02 *mk48t02)
{
    time_t     t;
    struct tm  *tm;

    t = time(NULL);
    tm = localtime(&t);

    mk48t02->sec  = bin2bcd(tm->tm_sec);
    mk48t02->min  = bin2bcd(tm->tm_min);
    mk48t02->hour = bin2bcd(tm->tm_hour);
    mk48t02->dw   = tm->tm_wday;
    mk48t02->day  = bin2bcd(tm->tm_mday);
    mk48t02->mon  = bin2bcd(tm->tm_mon + 1);
    mk48t02->year = bin2bcd(tm->tm_year - (2000 - 1900));
}



/*
 *  R e s e t
 */
int mk48t02Reset(Device *dev)
{
    if (!quiet)  printf(" MK48T02");

    return 0;
}


/*
 *  R e a d   o n e   b y t e
 */
static u_long mk48t02Read(Device *dev, u_long offset)
{
    MK48T02 *mk48t02 = (MK48T02 *) dev->priv;
    u_char  data;

    update(mk48t02);

    switch (offset & (REGSIZE-1))  {
        case CNTL :
            data = mk48t02->cntl;
            break;
        case SEC :
            data = mk48t02->sec;
            break;
        case MIN :
            data = mk48t02->min;
            break;
        case HOUR :
            data = mk48t02->hour;
            break;
        case DW :
            data = mk48t02->dw;
            break;
        case DAY :
            data = mk48t02->day;
            break;
        case MON :
            data = mk48t02->mon;
            break;
        case YEAR :
            data = mk48t02->year;
            break;
        default :
            if (offset & 0x0001)
                data = mk48t02->sram[(offset & (REGSIZE-1)) >> 1];
            else
                data = 0xFF;
            break;
    }


    return (u_long) data;
}


/*
 *  W r i t e   o n e   b y t e
 */
static void mk48t02Write(Device *dev, u_long offset, u_char data)
{
    MK48T02 *mk48t02 = (MK48T02 *) dev->priv;

    switch (offset & (REGSIZE-1))  {
        case CNTL :
            mk48t02->cntl = data;
            break;
        case SEC :
            mk48t02->sec = data;
            break;
        case MIN :
            mk48t02->min = data;
            break;
        case HOUR :
            mk48t02->hour = data;
            break;
        case DW :
            mk48t02->dw = data;
            break;
        case DAY :
            mk48t02->day = data;
            break;
        case MON :
            mk48t02->mon = data;
            break;
        case YEAR :
            mk48t02->year = data;
            break;
        default :
            if (offset & 0x0001)
                mk48t02->sram[(offset & (REGSIZE-1)) >> 1] = data;
            break;
    }
}


/*
 *  I n i t i a l i z e
 */
int initMK48T02(Device *dev, u_long base, u_long size, void *arg)
{
    MK48T02 *mk48t02;


    if ((mk48t02 = malloc(sizeof(MK48T02))) == NULL) {
        fprintf(stderr, "MK48T02: Malloc of %ld bytes failed - %s\n", size, strerror(errno));
        exit(errno);
    }

    dev->base = (u_long) base;
    dev->mask = ~(size-1);
    dev->priv = (void *) mk48t02;
    dev->devReset = mk48t02Reset;
    dev->devRead  = mk48t02Read;
    dev->devWrite = mk48t02Write;
    dev->devIAck  = NULL;
    dev->devCycle = NULL;

    return 0;
}

