/*
 * Most of this code was written by Marwan Shaban at North Carolina
 * University's Electrical and Computer Engineering department in 1989.
 *
 * Marwan Shaban, shaban@ece36h.ncsu.edu
 *
 * Ported to Linux as a part of the 68000 system simulator by
 * Peter Blum, Peter.Blum@readme1st.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <stdio.h>

#include "m68000.h"
#include "decoder.h"
#include "opcodes.h"



extern int   cpuState;          /* CPU state */
extern long  inst;              /* Current instruction */

/* Return of effAddr() */
static long  *EA1, *EA2;        /* Pointer to register memaddr or immdata */
static long  memaddr;           /* hold the memory address operand of EA */
static long  immdata;           /* hold an immediate data operand of EA */

static long  source, dest, result;


/*
 * Converts a number from 2's compliment notation to the "C" language
 * format so that operations may be performed on it.
 */
static void from_2s_comp(long number, long size, long *result)
{
    if (size == LONG)  {
        if (number & 0x80000000)  {
            *result = ~number + 1;
            *result = -*result;
        }
        else  {
            *result = number;
        }
    }
    else if (size == WORD) {
        if (number & 0x8000)  {
            *result = (~number + 1) & WORD;
            *result = -*result;
        }
	else  {
            *result = number;
        }
    }
    else  {
        if (number & 0x80)  {
            *result = (~number + 1) & BYTE;
            *result = -*result;
        }
        else {
            *result = number;
        }
    }
}


/*
 * sign-extends a number from byte or word to long
 */
static void signExtend(int number, long size, long *result)
{

    if (size == WORD)  {
        if (number & 0x8000)
            *result = number | 0xffff0000;
        else
            *result = number & 0x0000ffff;
    }
    else if (size == BYTE)  {
        if (number & 0x80)
            *result = number | 0xffffff00;
        else
            *result = number & 0x000000ff;
    }
    else  {
        *result = number & size;
    }
}



/*
 * Returns the value of the location referenced regardless of whether it is
 * a memory location or a 68000 register location.
 */
static void valueOf(long *EA, long *value, long size)
{
	if (EA == &memaddr)
		memGet(*EA, size, value);
	else
		*value = *EA & size;
}


/*
 * Decodes the size field in the instruction being processed
 * and returns a mask to be used in instruction execution.
 */
static int decodeSize(long *result)
{
	switch ((inst >> 6) & 0x0003)  {
	    case 0 : *result = BYTE; break;
	    case 1 : *result = WORD; break;
	    case 2 : *result = LONG; break;
	    default: *result = 0; return BAD_INST;
	}

	return SUCCESS;
}


/*
 * Decodes the effective address field in the current instruction,
 * returns a pointer to the effective address either in EA1 or EA2
 */
static int effAddr(long size, int mask)
{
    int   mode, reg, addr;
    int   move;
    int   legal = 0;
    long  ext, inc_size, ind_reg, disp;
    long  *value = NULL;


    if ((((inst & 0xf000) == 0x1000) || ((inst & 0xf000) == 0x2000) || ((inst & 0xf000) == 0x3000)) && (mask == DATA_ALT_ADDR))  {
        move = 1;
        /* reg and mode are reversed in MOVE dest EA */
        addr = (inst >> 6) & 0x003f;
        mode = addr & REG_MASK;
        reg  = (addr & MODE_MASK) >> 3;
    }
    else  {
        move = 0;
        addr = inst & 0x003f;
        mode = (addr & MODE_MASK) >> 3;
        reg  = addr & REG_MASK;
    }

    switch (mode) {
        case 0 :
            if (mask & bit_1)  {
                value = &D[reg];
                if (trace) printf("D%d", reg);
                legal = 1;
            }
            break;
        case 1 :
            if (mask & bit_2)  {
                reg = a_reg(reg);
                value = &A[reg];
                if (trace) printf("A%d", reg);
                legal = 1;
            }
            break;
        case 2 :
            if (mask & bit_3)  {
                reg = a_reg(reg);
                memaddr = A[reg];
                value = &memaddr;
                if (trace) printf("(A%d)", reg);
                legal = 1;
            }
            break;
        case 3 :
            if (mask & bit_4)  {
                reg = a_reg(reg);
                if (size == BYTE) 
                    inc_size = 1;
                else if (size == WORD)
                    inc_size = 2;
                else
                    inc_size = 4;
                memaddr = A[reg];
                value = &memaddr;
                A[reg] = A[reg] + inc_size;
                if (trace) printf("(A%d)+", reg);
                legal = 1;
            }
            break;
        case 4 :
            if (mask & bit_5)  {
                reg = a_reg(reg);
                if (size == BYTE) 
                    inc_size = 1;
                else if (size == WORD)
                    inc_size = 2;
                else
                    inc_size = 4;
                A[reg] = A[reg] - inc_size;
                memaddr = A[reg];
                value = &memaddr;
                if (trace) printf("-(A%d)", reg);
                legal = 1;
            }
            break;
        case 5 :
            if (mask & bit_6)  {
                reg = a_reg(reg);
                memRequest(&PC, WORD, &ext);
                from_2s_comp(ext, WORD, &ext);
                memaddr = A[reg] + ext;
                value = &memaddr;
                if (trace)  printf("%04x(A%d)", (unsigned short) ext, reg);
                legal = 1;
            }
            break;
        case 6 :
            if (mask & bit_7)  {
                reg = a_reg(reg);
                /* Get extension word */
                memRequest(&PC, WORD, &ext);
                disp = ext & 0x00ff;
                signExtend(disp, BYTE, &disp);
                from_2s_comp(disp, WORD, &disp);
                /* Get index register value */
                if (ext & 0x8000)
                    ind_reg = A[a_reg((ext & 0x7000) >> 12)];
                else
                    ind_reg = D[(ext & 0x7000) >> 12];
		/* Get correct length for index register */
                if (!(ext & 0x0800)) {
                    signExtend(ind_reg, WORD, &ind_reg);
                    from_2s_comp(ind_reg, LONG, &ind_reg);
                }
                memaddr = A[reg] + disp + ind_reg;
                value = &memaddr;
                if (trace)  printf("%02x(A%d, dx)", (unsigned char) disp, reg);
                legal = 1;
            }
            break;
        case 7 :
            switch (reg) {
                case 0 :
                    if (mask & bit_8) {
                        memRequest(&PC, WORD, &ext);
                        from_2s_comp(ext, WORD, &ext);
                        memaddr = ext;
                        value = &memaddr;
                        if (trace)  printf("%08lx.w", ext);
                        legal = 1;
                    }
                    break;
                case 1 :
                    if (mask & bit_9)  {
                        memRequest(&PC, LONG, &ext);
                        memaddr = ext;
                        value = &memaddr;
                        if (trace)  printf("%08lx.l", ext);
                        legal = 1;
                    }
                    break;
                case 2 :
                    if (mask & bit_10) {
                        memRequest(&PC, WORD, &ext);
                        from_2s_comp(ext, WORD, &ext);
                        if (trace)  printf("%04x(PC)", (unsigned short) ext);
                        memaddr = PC - 2  + ext;
                        value = &memaddr;
                        legal = 1;
                    }
                    break;
                case 3 :
                    if (mask & bit_11)  {
                        /* Get extension word */
                        memRequest(&PC, WORD, &ext);
                        disp = ext & 0xff;
                        signExtend(disp, BYTE, &disp);
                        from_2s_comp(disp, WORD, &disp);
                        if (trace)  printf("%02x(PC,Dx)", (unsigned char) disp);
                        /* Get index register value */
                        if (ext & 0x8000)
                            ind_reg = A[a_reg((ext & 0x7000) >> 12)];
                        else
                            ind_reg = D[(ext & 0x7000) >> 12];
                        /* Get correct length for index register */
                        if (!(ext & 0x0800)) {
                            signExtend(ind_reg, WORD, &ind_reg);
                            from_2s_comp(ind_reg, LONG, &ind_reg);
                        }
                        ext = ext & 0x00ff;
                        memaddr = PC - 2 + disp + ind_reg;
                        value = &memaddr;
                        legal = 1;
                    }
                    break;
                case 4 :
                    if (mask & bit_12)  {
                        if ((size == BYTE) || (size == WORD))
                            memRequest(&PC, WORD, &ext);
                        else
                            memRequest(&PC, LONG, &ext);
                        immdata = ext;
                        value = &immdata;
                        if (trace)  printf("#%lx", ext);
                        legal = 1;
                    }
                    break;
            }
            break;
    }
    if (!legal)  return BAD_INST;

    if (move)  {
        EA2 = value;
    }
    else {
        EA1 = value;
    }

    return SUCCESS;
}



/*
 * Performs the task of putting the result of some operation into a destination
 * location "according to size".  This means that the bits of the destination
 * that are in excess to the size of the operation are not affected.
 * This function provides a general-purpose mechanism for putting the result of
 * an operation in the destination, no matter whether the destination is
 * a memory location or a 68000 register.  The data is placed in the destination
 * correctly and the rest of the bits in a register are left alone.
 */
static void put(long *dst, long src, long size)
{
	if (dst == &memaddr)
		memPut(src, *dst, size);
	else
		*dst = (src & size) | (*dst & ~size);
}



/*
 * Updates the five condition codes according to the codes passed as parameters.
 * Each of the condition codes has a number of ways it can be calculated, and
 * the appropriate method of computation is passed as the parameter for that
 * condition code.
 * The src, dst, and res operands contain the source, destination, and result
 * of the instruction requesting updating the condition codes.
 * Also, for shift and rotate instructions the shift count needs to be passed.
 */
static int ccrUpdate(int x, int n, int z, int v, int c, long source, long dest, long result, long size, int r)
{
	int	x_bit, n_bit, z_bit, v_bit, c_bit;
	long	Rm, Dm, Sm;
	long	m;

	source &= size;
	dest &= size;
	result &= size;

	if (size == BYTE)  {
		m = 7;
		Sm = source & 0x0080;
		Rm = result & 0x0080;
		Dm = dest & 0x0080;
	}
	else if (size == WORD)  {
		m = 15;
		Sm = source & 0x8000;
		Rm = result & 0x8000;
		Dm = dest & 0x8000;
	}
	else  {
		m = 31;
		Sm = source & 0x80000000;
		Rm = result & 0x80000000;
		Dm = dest & 0x80000000;
	}

	/*
	 * Calculate each of the five condition codes according to the requested
	 * method of calculation
	 */
	switch (n) {
	    case GEN :
		n_bit = (Rm) ? nbit : 0;
		break;
	    default :
		n_bit = SR & nbit;
		break;
	}

	switch (z) {
	    case GEN :
		z_bit = (result == 0) ? zbit : 0;
		break;
	    case CASE_1 :
		z_bit = ((SR & zbit) && (result == 0)) ? zbit : 0;
		break;
	    default :
		z_bit = SR & zbit;
		break;
	}

	switch (v) {
	    case ZER :
		v_bit = 0;
		break;
	    case CASE_1 :
		v_bit = ((Sm && Dm && !Rm) || (!Sm && !Dm && Rm)) ? vbit : 0;
		break;
	    case CASE_2 :
		v_bit = ((!Sm && Dm && !Rm) || (Sm && !Dm && Rm)) ?  vbit : 0;
		break;
	    case CASE_3 :
		v_bit = (Dm && Rm) ? vbit : 0;
		break;
	    default :
		v_bit = SR & vbit;
		break;
	}

	switch (c) {
	    case ZER :
		c_bit = 0;
		break;
	    case CASE_1 :
		c_bit = (SR & xbit) ? cbit : 0;
		break;
	    case CASE_2 :
		c_bit = ((dest >> (r-1)) & 1) ? cbit : 0;
		break;
	    case CASE_3 :
		c_bit = ((dest >> (m-r+1)) & 1) ? cbit : 0;
		break;
	    case CASE_4 :
		c_bit = (Dm || Rm) ? cbit : 0;
		break;
	    case CASE_5 :
		c_bit = ((Sm && Dm) || (!Rm && Dm) || (Sm && !Rm)) ?  cbit : 0;
		break;
	    case CASE_6 :
		c_bit = ((Sm && !Dm) || (Rm && !Dm) || (Sm && Rm)) ?  cbit : 0;
		break;
	    default :
		c_bit = SR & cbit;
		break;
	}

	switch (x) {
	    case GEN :
		x_bit = c_bit ? xbit : 0;
		break;
	    default :
		x_bit = SR & xbit;
		break;
	}

	/* set SR according to results */
	SR = SR & 0xffe0;
	SR = SR | (x_bit | n_bit | z_bit | v_bit | c_bit);

	return SUCCESS;
}


/*
 * Checks for the truth of a certain condition and returns the result.
 */
int checkCondition(int condition)
{
	int   result = 0;

	switch (condition)  {
	    case T :
		if (trace)  printf("t");
		result = 1;				/* true */
		break;

	    case F :
		if (trace)  printf("f");
		result = 0;				/* false */
		break;

	    case HI :
		if (trace)  printf("hi");
		result = !(SR & cbit) && !(SR & zbit);	/* high */
		break;

	    case LS :
		if (trace)  printf("ls");
		result = (SR & cbit) || (SR & zbit);	/* low or same */
		break;

	    case CC :
		if (trace)  printf("cc");
		result = !(SR & cbit);			/* carry clear */
		break;

	    case CS :
		if (trace)  printf("cs");
		result = (SR & cbit);			/* carry set */
		break;

	    case NE :
		if (trace)  printf("ne");
		result = !(SR & zbit);			/* not equal */
		break;

	    case EQ :
		if (trace)  printf("eq");
		result = (SR & zbit);			/* equal */
		break;

	    case VC :
		if (trace)  printf("vc");
		result = !(SR & vbit);			/* overflow clear */
		break;

	    case VS :
		if (trace)  printf("vs");
		result = (SR & vbit);			/* overflow set */
		break;

	    case PL :
		if (trace)  printf("pl");
		result = !(SR & nbit);			/* plus */
		break;

	    case MI :
		if (trace)  printf("mi");
		result = (SR & nbit);			/* minus */
		break;

	    case GE :
		if (trace)  printf("ge");
		result = ((SR & nbit) && (SR & vbit)) ||
	        	(!(SR & nbit) && !(SR & vbit));
		break;

	    case LT :
		if (trace)  printf("lt");
		result = ((SR & nbit) && !(SR & vbit)) ||
			(!(SR & nbit) && (SR & vbit));
		break;

	    case GT :
		if (trace)  printf("gt");
		result = ((SR & nbit) && (SR & vbit) && !(SR & zbit)) ||
			(!(SR & nbit) && !(SR & vbit) && !(SR & zbit));
		break;

	    case LE :
		if (trace)  printf("le");
		result = ((SR & nbit) && !(SR & vbit)) ||
			(!(SR & nbit) && (SR & vbit)) || (SR & zbit);
		break;
	}

	return result;
}


int MOVE(void)
{
    long  size;
    long  src, dst;

    switch ((inst >> 12) & 0x03)  {
        case 0x01 : size = BYTE; break;		/* bit pattern '01' */
        case 0x03 : size = WORD; break;		/* bit pattern '10' */
        case 0x02 : size = LONG; break;		/* bit pattern '11' */
        default   : return BAD_INST;		/* bad size field */
    }

    if (effAddr(size, ALL_ADDR))  return BAD_INST;
    valueOf(EA1, &src, size);

    if (trace)  printf(",");

    if (effAddr(size, DATA_ALT_ADDR))  return BAD_INST;
    valueOf(EA2, &dst, size);

    dst = src;
    put(EA2, dst, size);
    ccrUpdate(N_A, GEN, GEN, ZER, ZER, src, dst, dst, size, 0);

    return SUCCESS;
}


int MOVEP(void)
{
    int   address, Dx, disp, count, direction, reg;
    long  temp, size;

    memRequest(&PC, WORD, &temp);
    from_2s_comp(temp, (long) WORD, &temp);

    address = A[a_reg(inst & 0x0007)] + (temp & WORD);
    direction = inst & 0x80;

    if (inst & 0x40)  {
        size = LONG;
        count = 4;
    }
    else  {
        size = WORD;
        count = 2;
    }

    reg = (inst >> 9) & 0x07;
    Dx = D[reg] & size;

    for (;count > 0; count--)  {
        disp = 8 * (count - 1);
        if (direction)  {
            memPut((long)((Dx >> disp) & BYTE) , address, BYTE);
        }
        else  {
            memGet(address, BYTE, &temp);
            switch (count) {
              case 4 :
                D[reg] = (D[reg] & 0x00ffffff) | (temp * 0x1000000);
                break;
              case 3 :
                D[reg] = (D[reg] & 0xff00ffff) | (temp * 0x10000);
                break;
              case 2 :
                D[reg] = (D[reg] & 0xffff00ff) | (temp * 0x100);
                break;
              case 1 :
                D[reg] = (D[reg] & 0xffffff00) | (temp);
                break;
            }
        }
        address += 2;
    }

    return SUCCESS;
}


int MOVEA(void)
{
    long  size;
    int   reg;

    size = (inst & 0x1000) ? WORD : LONG;

    if (effAddr(size, ALL_ADDR))  return BAD_INST;

    valueOf(EA1, &source, size);
    if (size == WORD)
        signExtend(source, WORD, &source);

    reg = (inst >> 9) & 0x07;
    A[a_reg(reg)] = source;

    if (trace)  printf(",A%d", reg);

    return SUCCESS;
}


int MOVE_SR_(void)
{
    if (trace)  printf("SR,");

    if (effAddr(WORD, DATA_ALT_ADDR))  return BAD_INST;

    put(EA1, SR, WORD);

    return SUCCESS;
}


int MOVE_CCR(void)
{
    long  temp;

    if (effAddr(WORD, DATA_ADDR))  return BAD_INST;

    valueOf(EA1, &temp, WORD);
    put(&SR, temp, BYTE);

    if (trace)  printf(",CCR");

    return SUCCESS;
}


int MOVE_SR(void)
{
    long  temp;

    if (!(SR & sbit))
        return NO_PRIVILEGE;

    if (effAddr(WORD, DATA_ADDR))  return BAD_INST;

    valueOf(EA1, &temp, WORD);
    put(&SR, temp, WORD);

    if (trace)  printf(",SR");

    return SUCCESS;
}


int MOVEM(void)
{
    int   direction, addr_modes_mask, cnt;
    int   displacement;
    long  address;
    long  size, mask_list, temp;

    memRequest(&PC, WORD, &mask_list);

    size = (inst & 0x0040) ? LONG : WORD;
    direction = (inst & 0x0400);
    if (direction)
        addr_modes_mask = CONTROL_ADDR | bit_4;
    else
        addr_modes_mask = CONT_ALT_ADDR | bit_5;

    if (effAddr(size, addr_modes_mask))  return BAD_INST;

    address = *EA1;

    if ((inst & 0x0038) != 0x20)  {
        if (size == WORD)
            displacement = 2;
        else
            displacement = 4;
    }
    else  {
        if (size == WORD)
            displacement = -2;
        else
            displacement = -4;
    }

    for (cnt = 0; cnt < 16; cnt++)  {
        if (mask_list & (1 << cnt)) {
            if (direction) {
                memGet(address, size, &temp);
                if (size == WORD)
                    signExtend(temp, LONG, &temp);
                if (cnt < 8)
                    D[cnt] = temp;
                else
                    A[a_reg(cnt - 8)] = temp;
            }
            else  {
                if ((inst & 0x0038) == 0x0020) {
                    if (cnt < 8)
                        memPut(A[a_reg(7 - cnt)], address, size);
                    else
                        memPut(D[15 - cnt], address, size);
                }
                else  {
                    if (cnt < 8)
                        memPut(D[cnt], address, size);
                    else
                        memPut(A[a_reg(cnt - 8)], address, size);
                }
            }
            address += displacement;
        }
    }

    if ((inst & 0x0038) == 0x0020)  {
        address -= displacement;
        A[a_reg(inst & 0x7)] = address;
    }
    else if ((inst & 0x0038) == 0x0018)  {
        A[a_reg(inst & 0x7)] = address;
    }

    return SUCCESS;
}


int MOVE_USP(void)
{
    int   regNum;

    if (!(SR & sbit))
        return NO_PRIVILEGE;

    regNum = a_reg(inst & 0x0007);

    if (inst & 0x0008) {
        A[regNum] = A[7];
        if (trace)  printf("USP,a%d", regNum);
    }
    else  {
        A[7] = A[regNum];
        if (trace)  printf("a%d,USP", regNum);
    }

    return SUCCESS;
}


int MOVEQ(void)
{
    int   reg;

    reg = (inst >> 9) & 0x7;
    source = inst & 0x00ff;

    signExtend(source, BYTE, &source);
    signExtend(source, WORD, &source);

    D[reg] = dest = source;

    ccrUpdate(N_A, GEN, GEN, ZER, ZER, source, dest, D[reg], LONG, 0);

    if (trace)  printf("#%ld,D%d", source, reg);

    return SUCCESS;
}


int EXG(void)
{
    long  temp_reg;
    int   Rx, Ry;

    Rx = (inst >> 9) & 0x07;
    Ry = inst & 0x0007;

    switch ((inst >> 3) & 0x1f) {
      case 0x08:
        temp_reg = D[Rx];
        D[Rx] = D[Ry];
        D[Ry] = temp_reg;
        break;
      case 0x09:
        temp_reg = A[a_reg(Rx)];
        A[a_reg(Rx)] = A[a_reg(Ry)];
        A[a_reg(Ry)] = temp_reg;
        break;
      case 0x11:
        temp_reg = D[Rx];
        D[Rx] = A[a_reg(Ry)];
        A[a_reg(Ry)] = temp_reg;
        break;
      default :
        return BAD_INST;
        break;
    }

    return SUCCESS;
}


int LEA(void)
{
    int   reg;

    reg = (inst >> 9) & 0x07;

    if (effAddr(LONG, CONTROL_ADDR))  return BAD_INST;
    A[a_reg(reg)] = *EA1;

    if (trace)  printf(",A%d", reg);

    return SUCCESS;
}


int PEA(void)
{
    if (effAddr(LONG, CONTROL_ADDR))  return BAD_INST;

    A[a_reg(7)] -= 4;
    memPut(*EA1 , A[a_reg(7)], LONG);

    return SUCCESS;
}


int LINK(void)
{
    int   reg;
    long  temp, disp;

    reg = inst & 0x0007;

    memRequest(&PC, WORD, &temp);
    from_2s_comp(temp, WORD, &disp);

    A[a_reg(7)] -= 4;
    memPut(A[reg], A[a_reg(7)], LONG);
    A[reg] = A[a_reg(7)];
    A[a_reg(7)] = A[a_reg(7)] + disp;

    if (trace)  printf("#%04x", (unsigned short) temp);

    return SUCCESS;
}


int UNLK(void)
{
    int   reg;

    reg = inst & 0x0007;

    A[a_reg(7)] = A[reg];
    memGet(A[a_reg(7)], LONG, &A[reg]);
    A[a_reg(7)] += 4;

    return SUCCESS;
}


int ADD(void)
{
    int	  addr_modes_mask;
    int	  reg;
    long  size;

    addr_modes_mask = (inst & 0x0100) ? MEM_ALT_ADDR : ALL_ADDR;

    if (decodeSize(&size))  return BAD_INST;
    if (effAddr(size, addr_modes_mask))  return BAD_INST;

    reg = (inst >> 9) & 0x0007;	/* get data register number */

    if (inst & 0x0100) {
        source = D[reg];
        valueOf(EA1, &dest, size);
        result = source + dest;
        put(EA1, result, size);
    }
    else {
        valueOf(EA1, &source, size);
        dest = D[reg];
        result = source + dest;
        put(&D[reg], result, size);
    }
    ccrUpdate(GEN, GEN, GEN, CASE_1, CASE_5, source, dest, result, size, 0);

    return SUCCESS;
}


int ADDA(void)
{
    long  size;
    int   reg;

    size = (inst & 0x0100) ? LONG : WORD;

    if (effAddr(size, ALL_ADDR))  return BAD_INST;

    reg = (inst >> 9) & 0x0007;

    valueOf(EA1, &source, size);
    dest = A[a_reg(reg)];
    put(&A[a_reg(reg)], source + dest, size);

    if (trace)  printf(",A%d", reg);

    return SUCCESS;
}


int ADDI(void)
{
    long  size;

    if (decodeSize(&size))  return BAD_INST;

    memRequest(&PC, size, &source);

    if (trace)  printf("#%lx,", source);

    if (effAddr(size, DATA_ALT_ADDR))  return BAD_INST;

    valueOf(EA1, &dest, size);
    result = source + dest;
    put(EA1, result, size);
    ccrUpdate(GEN, GEN, GEN, CASE_1, CASE_5, source, dest, result, size, 0);

    return SUCCESS;
}


int ADDQ(void)
{
    long  size;

    if (decodeSize(&size))  return BAD_INST;

    if ((inst & 0x0038) == 0x0008)
        /* if address reg direct, operation is long */
        size = LONG;

    source = ((inst >> 9) & 0x07);
    if (source == 0)  source = 8;

    if (trace)  printf("#%ld,", source);

    if (effAddr(size, ALTERABLE_ADDR))  return BAD_INST;

    valueOf(EA1, &dest, size);
    result = source + dest;
    put(EA1, result, size);

    if (!((inst & 0x0038) == 0x0008))
        /* Address reg direct: cc's not affected */
        ccrUpdate(GEN, GEN, GEN, CASE_1, CASE_5, source, dest, result, size, 0);

    return SUCCESS;
}


int ADDX(void)
{
    long  size;
    int   Rx, Ry;

    if (decodeSize(&size))  return BAD_INST;

    Rx = (inst >> 9) & 0x0007;
    Ry = inst & 0x0007;

    if (inst & 0x0008)  {
        Rx = a_reg(Rx);
        Ry = a_reg(Ry);
        if (size == BYTE) {
            A[Rx]--;
            A[Ry]--;
        }
        else if (size == WORD) {
            A[Rx] -= 2;
            A[Ry] -= 2;
        }
        else {
            A[Rx] -= 4;
            A[Ry] -= 4;
        }
        memGet(A[Ry], size, &source);
        memGet(A[Rx], size, &dest);
        memPut(A[a_reg(Rx)], source + dest + ((SR & xbit) >> 4), size);
        memGet(A[Rx], size, &result);
    }
    else  {
        source = D[Ry] & size;
        dest = D[Rx] & size;
        put(&D[Rx], source + dest + ((SR & xbit) >> 4), size);
        result = D[Rx] & size;
    }
    ccrUpdate(GEN, GEN, CASE_1, CASE_1, CASE_5, source, dest, result, size, 0);

    return SUCCESS;
}


int SUB(void)
{
    int   addr_modes_mask, reg;
    long  size;

    addr_modes_mask = (inst & 0x0100) ? MEM_ALT_ADDR : ALL_ADDR;

    if (decodeSize(&size))  return BAD_INST;
    if (effAddr(size, addr_modes_mask))  return BAD_INST;

    reg = (inst >> 9) & 0x0007;

    if (inst & 0x0100) {
        source = D[reg] & size;
        valueOf(EA1, &dest, size);
        result = dest - source;
        put(EA1, result, size);
    }
    else  {
        valueOf(EA1, &source, size);
        dest = D[reg] & size;
        result = dest - source;
        put(&D[reg], result, size);
    }
    ccrUpdate(GEN, GEN, GEN, CASE_2, CASE_6, source, dest, result, size, 0);

    if (trace)  printf(",D%d", reg);

    return SUCCESS;
}


int SUBA(void)
{
    long  size;
    int   reg;

    size = (inst & 0x0100) ? LONG : WORD;

    if (effAddr(size, ALL_ADDR))  return BAD_INST;

    reg = (inst >> 9) & 0x0007;

    valueOf(EA1, &source, size);
    dest = A[a_reg(reg)];
    put(&A[a_reg(reg)], dest - source, size);

    if (trace)  printf(",A%d", reg);

    return SUCCESS;
}


int SUBI(void)
{
    long  size;

    if (decodeSize(&size))  return BAD_INST;

    memRequest(&PC, size, &source);

    if (trace)  printf("#%lx,", source);

    if (effAddr(size, DATA_ALT_ADDR))  return BAD_INST;

    valueOf(EA1, &dest, size);
    result = dest - source;
    put(EA1, result, size);
    ccrUpdate(GEN, GEN, GEN, CASE_2, CASE_6, source, dest, result, size, 0);

    return SUCCESS;
}


int SUBQ(void)
{
    long  size;

    if (decodeSize(&size))  return BAD_INST;

    if ((inst & 0x0038) == 0x0008)
        size = LONG;

    source = ((inst >> 9) & 0x07);
    if (source == 0)  source = 8;

    if (trace)  printf("#%ld,", source);

    if (effAddr(size, ALTERABLE_ADDR))  return BAD_INST;

    valueOf(EA1, &dest, size);
    result = dest - source;
    put(EA1, result, size);

    if (!((inst & 0x0038) == 0x0008))
        ccrUpdate(GEN, GEN, GEN, CASE_2, CASE_6, source, dest, result, size, 0);

    return SUCCESS;
}


int SUBX(void)
{
    long  size;
    int   Rx, Ry;

    if (decodeSize(&size))  return BAD_INST;

    Ry = (inst >> 9) & 0x0007;
    Rx = inst & 0x0007;

    if (inst & 0x0008)  {
        Rx = a_reg(Rx);
        Ry = a_reg(Ry);
        if (size == LONG)  {
            A[Rx] -= 4;
            A[Ry] -= 4;
        }
        else if (size == WORD)  {
            A[Rx] -= 2;
            A[Ry] -= 2;
        }
        else  {
            A[Rx] -= 1;
            A[Ry] -= 1;
        }
        memGet(A[Rx], size, &source);
        memGet(A[Ry], size, &dest);
        put((long *)A[Ry], dest - source - ((SR & xbit)>> 4), size);
        memGet(A[Ry], size, &result);
    }
    else  {
        source = D[Rx] & size;
        dest = D[Ry] & size;
        put(&D[Ry], dest - source - ((SR & xbit) >> 4), size);
        result = D[Ry] & size;
    }

    ccrUpdate(GEN, GEN, CASE_1, CASE_2, CASE_6, source, dest, result, size, 0);

    return SUCCESS;
}


int DIVS(void)
{
    int   reg, overflow;
    long  remainder;

    if (effAddr(WORD, DATA_ADDR))  return BAD_INST;

    reg = (inst >> 9) & 0x0007;

    valueOf(EA1, &source, WORD);

    from_2s_comp(source, WORD, &source);
    from_2s_comp(D[reg], LONG, &dest);

    if (source == 0)  return DIV_BY_ZERO;

    /* Check for overflow */
    if (((source < 0) && (dest >= 0)) || ((source >= 0) && (dest < 0))) {
        if ((dest / -source) > WORD)
            overflow = 1;
        else
            overflow = 0;
    }
    else  {
        if ((dest / source) > WORD)
            overflow = 1;
        else
            overflow = 0;
    }

    if (overflow)
        SR |= vbit;
    else  {
        SR &= ~vbit;
        result = (dest / source) & WORD;
        remainder = (dest % source) & WORD;
        D[reg] = result = result | (remainder * 0x10000);
        ccrUpdate(N_A, GEN, GEN, N_A, ZER, source, dest, result, (long) WORD, 0);
        if (((source < 0) && (dest >= 0)) || ((source >= 0) && (dest < 0)))
            SR |= nbit;
        else
            SR &= ~nbit;
    }

    return SUCCESS;
}


int DIVU(void)
{
    int   reg;

    if (effAddr(WORD, DATA_ADDR))  return BAD_INST;

    reg = (inst >> 9) & 0x0007;

    valueOf(EA1, &source, WORD);
    dest = D[reg];

    if (source == 0)  return DIV_BY_ZERO;

    if ((dest / source) > WORD)  {
        SR |= vbit;
    }
    else  {
        SR &= ~vbit;
        D[reg] = result = ((dest / source) & 0xffff) | ((dest % source) * 0x10000);
        ccrUpdate(N_A, GEN, GEN, N_A, ZER, source, dest, result, WORD, 0);
    }

    return SUCCESS;
}


int MULS(void)
{
    int   reg;

    if (effAddr(WORD, DATA_ADDR))  return BAD_INST;

    reg = (inst >> 9) & 0x0007;

    valueOf(EA1, &source, WORD);
    from_2s_comp(source, WORD, &source);
    from_2s_comp(D[reg] & WORD, WORD, &dest);

    D[reg] = result = source * dest;

    ccrUpdate(N_A, GEN, GEN, CASE_9, ZER, source, dest, result, WORD, 0);

    return SUCCESS;
}


int MULU(void)
{
    int   reg;

    if (effAddr(WORD, DATA_ADDR))  return BAD_INST;

    reg = (inst >> 9) & 0x0007;

    valueOf(EA1, &source, WORD);
    dest = D[reg] & WORD;

    D[reg] = result = source * dest;

    ccrUpdate(N_A, GEN, GEN, ZER, ZER, source, dest, result, LONG, 0);

    return SUCCESS;
}


int NEG(void)
{
    long  size;

    if (effAddr (size, DATA_ALT_ADDR))  return BAD_INST;
    if (decodeSize(&size))  return BAD_INST;

    valueOf(EA1, &source, size);
    result = -source;
    put(EA1, result, size);

    ccrUpdate(GEN, GEN, GEN, CASE_3, CASE_4, source, dest, result, size, 0);

    return SUCCESS;
}


int NEGX(void)
{
    long  size;

    if (effAddr(size, DATA_ALT_ADDR))  return BAD_INST;
    if (decodeSize(&size))  return BAD_INST;

    valueOf(EA1, &source, size);
    result = -source - ((SR &xbit) >> 4);
    put(EA1, result, size);

    ccrUpdate(GEN, GEN, CASE_1, CASE_3, CASE_4, source, dest, result, size, 0);

    return SUCCESS;
}


int CMP(void)
{
    long  size;
    int   reg;

    if (decodeSize(&size))  return BAD_INST;
    if (effAddr(size, ALL_ADDR))  return BAD_INST;

    reg = (inst >> 9) & 0x0007;
    valueOf(EA1, &source, size);
    dest = D[reg] & size;
    result = dest - source;
    ccrUpdate(N_A, GEN, GEN, CASE_2, CASE_6, source, dest, result, size, 0);

    if (trace)  printf(",D%d", reg);

    return SUCCESS;
}


int CMPA(void)
{
    long  size;
    int   reg;

    size = (inst & 0x0100) ? LONG : WORD;

    if (effAddr(size, ALL_ADDR))  return BAD_INST;

    reg = a_reg((inst >> 9) & 0x0007);

    valueOf(EA1, &source, size);
    dest = A[reg] & size;
    if (size == WORD)
        signExtend(source, WORD, &source);
    result = dest - source;

    ccrUpdate(N_A, GEN, GEN, CASE_2, CASE_6, source, dest, result, size, 0);

    if (trace)  printf(",A%d", reg);

    return SUCCESS;
}


int CMPI(void)
{
    long  size;

    if (decodeSize(&size))  return BAD_INST;

    memRequest(&PC, size, &source);

    if (trace)  printf("#%lx,", source);

    if (effAddr(size, DATA_ALT_ADDR))  return BAD_INST;

    valueOf(EA1, &dest, size);
    result = dest - source;
    ccrUpdate(N_A, GEN, GEN, CASE_2, CASE_6, source, dest, result, size, 0);

    return SUCCESS;
}


int CMPM(void)
{
    long  size;
    int   Rx, Ry;

    if (decodeSize(&size))  return BAD_INST;

    Rx = a_reg((inst >> 9) & 0x07);
    Ry = a_reg(inst & 0x07);

    memGet(A[Ry], size, &source);
    memGet(A[Rx], size, &dest);

    result = dest - source;

    if (size == BYTE)  {
        A[Rx] += 1;
        A[Ry] += 1;
    }
    else if (size == WORD)  {
        A[Rx] += 2;
        A[Ry] += 2;
    }
    else  {
        A[Rx] += 4;
        A[Ry] += 4;
    }
    ccrUpdate(N_A, GEN, GEN, CASE_2, CASE_6, source, dest, result, size, 0);

    return SUCCESS;
}


int TST(void)
{
    long  size;

    if (decodeSize(&size))  return BAD_INST;
    if (effAddr(size, DATA_ALT_ADDR))  return BAD_INST;

    valueOf(EA1, &dest, size);
    ccrUpdate(N_A, GEN, GEN, ZER, ZER, source, dest, dest, size, 0);

    return SUCCESS;
}


int CLR(void)
{
    long  size;

    if (decodeSize(&size))  return BAD_INST;
    if (effAddr(size, DATA_ALT_ADDR))  return BAD_INST;

    valueOf(EA1, &dest, size);
    source = dest;
    put(EA1, 0, size);
    ccrUpdate(N_A, ZER, GEN, ZER, ZER, source, dest, 0, size, 0);

    return SUCCESS;
}


int EXT(void)
{
    long  size;
    int   reg;

    reg = inst & 0x07;

    size = (inst & 0x0040) ? LONG : WORD;
    source = dest = D[reg] & size;

    if (size == WORD)
        D[reg] = (D[reg]&~WORD) | (D[reg] & BYTE) | (0xff00 * ((D[reg]>>7) & 0x01));
    else
        D[reg] = (D[reg] & WORD) | (0xffff0000 * ((D[reg] >> 15) & 0x01));
    ccrUpdate(N_A, GEN, GEN, ZER, ZER, source, dest, D[reg], size, 0);

    return SUCCESS;
}


int ABCD(void)
{
    int   Rx, Ry, carry, temp_result;

    Rx = (inst >> 9) & 0x0007;
    Ry = inst & 0x0007;

    if (inst & 0x0008)  {
        /* Rx & Ry are address registers used in predecrement mode */
        Rx = a_reg(Rx);
        Ry = a_reg(Ry);
        A[Rx]--;
        A[Ry]--;
        memGet(A[Ry], BYTE, &source);
        memGet(A[Rx], BYTE, &dest);
    }
    else {
        /* Rx & Ry are data registers */
        source = D[Ry] & BYTE;
        dest = D[Rx] & BYTE;
    }

    result = ((SR & xbit) >> 4) + (source & 0xf) + (dest & 0xf);
    if (result > 9)  {
        result = result - 10;
        carry = 1;
    }
    else
        carry = 0;
    temp_result = ((source >> 4) & 0xf) + ((dest >> 4) & 0xf) + carry;
    if (temp_result > 9)  {
        temp_result = temp_result - 10;
        carry = 1;
    }
    else
        carry = 0;
    result = result + (temp_result << 4);

    if (inst & 0x0008)
        memPut(A[Rx], result, BYTE);
    else
        put(&D[Rx], result, BYTE);
    if (carry)
        SR = SR | cbit;
    else
        SR = SR & ~cbit;

    ccrUpdate(GEN, UND, CASE_1, UND, N_A, source, dest, result, (long) BYTE, 0);

    return SUCCESS;
}


int SBCD(void)
{
    int   Rx, Ry, borrow, temp_result;

    Rx = (inst >> 9) & 0x0007;
    Ry = inst & 0x0007;

    if (inst & 0x0008)  {
        /* Rx & Ry are address registers used in predecrement mode */
        Rx = a_reg(Rx);
        Ry = a_reg(Ry);
        A[Rx]--;
        A[Ry]--;
        memGet(A[Ry], BYTE, &source);
        memGet(A[Rx], BYTE, &dest);
    }
    else {
        /* Rx & Ry are data registers */
        source = D[Ry];
        dest = D[Rx];
    }

    result = (dest & 0xf) - (source & 0xf) - ((SR & xbit) >> 4);
    if (result < 0)  {
        result = result + 10;
        borrow = 1;
    }
    else
        borrow = 0;
    temp_result = ((dest >> 4) & 0xf) - ((source >> 4) & 0xf) - borrow;
    if (temp_result < 0)  {
        temp_result = temp_result + 10;
        borrow = 1;
    }
    else
        borrow = 0;
    result = result + (temp_result << 4);

    if (inst & 0x0008)
        memPut(A[Rx], result, BYTE);
    else
        put(&D[Rx], result, BYTE);

    if (borrow)
        SR = SR | cbit;
    else
        SR = SR & ~cbit;

    ccrUpdate(GEN, UND, CASE_1, UND, N_A, source, dest, result, (long) BYTE, 0);

    return SUCCESS;
}


int NBCD(void)
{
    int   borrow, temp_result;

    if (effAddr ((long) BYTE, DATA_ALT_ADDR))  return BAD_INST;

    valueOf(EA1, &dest, BYTE);
    result = 10 - (dest & 0xf) - ((SR & xbit) >> 4);

    if (result < 10)
        borrow = 1;
    else
        borrow = 0;

    temp_result = 10 - ((dest >> 4) & 0xf) - borrow;

    if (temp_result < 10)
        borrow = 1;
    else
        borrow = 0;

    result = result + (temp_result << 4);

    if (borrow)
        SR = SR | cbit;
    else
        SR = SR & ~cbit;

    put(EA1, result, BYTE);
    ccrUpdate(GEN, UND, CASE_1, UND, N_A, source, dest, result, (long) BYTE, 0);

    return SUCCESS;
}


int AND(void)
{
    int   addr_modes_mask, reg;
    long  size;

    addr_modes_mask = (inst & 0x0100) ? MEM_ALT_ADDR : DATA_ADDR;

    if (decodeSize(&size))  return BAD_INST;
    if (effAddr(size, addr_modes_mask))  return BAD_INST;

    reg = (inst >> 9) & 0x0007;

    if (inst & 0x0100) {
        source = D[reg] & size;
        valueOf(EA1, &dest, size);
        result = dest & source;
        put(EA1, result, size);
    }
    else {
        valueOf(EA1, &source, size);
        dest = D[reg] & size;
        result = dest & source;
        put(&D[reg], result, size);
    }
    ccrUpdate(N_A, GEN, GEN, ZER, ZER, source, dest, result, size, 0);

    return SUCCESS;
}


int ANDI(void)
{
    long  size;

    if (decodeSize(&size))  return BAD_INST;

    memRequest(&PC, size, &source);

    if (trace)  printf("#%lx,", source);

    if (effAddr(size, DATA_ALT_ADDR))  return BAD_INST;

    valueOf(EA1, &dest, size);
    result = source & dest;
    put(EA1, result, size);
    ccrUpdate(N_A, GEN, GEN, ZER, ZER, source, dest, result, size, 0);

    return SUCCESS;
}


int ANDI_CCR(void)
{
    long  temp;

    memRequest(&PC, WORD, &temp);

    SR &= (temp | 0xff00);

    if (trace)  printf("#%02x,CCR", (unsigned char) temp);

    return SUCCESS;
}


int ANDI_SR(void)
{
    long  temp;

    if (!(SR & sbit))
        return NO_PRIVILEGE;

    memRequest(&PC, WORD, &temp);
    SR &= temp;

    if (trace)  printf("#%04x,", (unsigned short) temp);

    return SUCCESS_PRIV;
}


int OR(void)
{
    long  size;
    int   mask, reg;

    mask = (inst & 0x0100) ? MEM_ALT_ADDR : DATA_ADDR;

    if (decodeSize(&size))  return BAD_INST;
    if (effAddr(size, mask))  return BAD_INST;

    reg = (inst >> 9) & 0x0007;

    if (inst & 0x0100) {
        source = D[reg] & size;
        valueOf(EA1, &dest, size);
        result = source | dest;
        put(EA1, result, size);
    }
    else {
        valueOf(EA1, &source, size);
        dest = D[reg] & size;
        result = source | dest;
        put(&D[reg], result, size);
    }
    ccrUpdate(N_A, GEN, GEN, ZER, ZER, source, dest, result, size, 0);

    return SUCCESS;
}


int ORI(void)
{
    long  size;

    if (decodeSize(&size))  return BAD_INST;

    memRequest(&PC, size, &source);

    if (trace)  printf("#%lx,", source);

    if (effAddr(size, DATA_ALT_ADDR))  return BAD_INST;

    valueOf(EA1, &dest, size);
    result = source | dest;
    put(EA1, result, size);
    ccrUpdate(N_A, GEN, GEN, ZER, ZER, source, dest, result, size, 0);

    return SUCCESS;
}


int ORI_CCR(void)
{
    long  temp;

    memRequest(&PC, WORD, &temp);
    SR |= (temp & WORD);

    if (trace)  printf("%02x,CCR", (unsigned char) temp);

    return SUCCESS;
}


int ORI_SR(void)
{
    long  temp;

    if (!(SR & sbit))
        return NO_PRIVILEGE;

    memRequest(&PC, WORD, &temp);
    SR |= temp;

    if (trace)  printf("%04x,SR", (unsigned short) temp);

    return SUCCESS_PRIV;
}


int EOR(void)
{
    long  size;
    int   reg;

    if (decodeSize(&size))  return BAD_INST;
    if (effAddr(size, DATA_ALT_ADDR))  return BAD_INST;

    reg = (inst >> 9) & 0x0007;

    source = D[reg] & size;
    valueOf(EA1, &dest, size);
    result = source ^ dest;
    put(EA1, result, size);
    ccrUpdate(N_A, GEN, GEN, ZER, ZER, source, dest, result, size, 0);

    return SUCCESS;
}


int EORI(void)
{
    long  size;

    if (decodeSize(&size))  return BAD_INST;

    memRequest(&PC, size, &source);

    if (trace)  printf("#%lx,", source);

    if (effAddr(size, DATA_ALT_ADDR))  return BAD_INST;

    valueOf(EA1, &dest, size);
    result = source ^ dest;
    put(EA1, result, size);
    ccrUpdate(N_A, GEN, GEN, ZER, ZER, source, dest, result, size, 0);

    return SUCCESS;
}


int EORI_CCR(void)
{
    long  temp;

    memRequest(&PC, WORD, &temp);

    SR ^= (temp & WORD);

    if (trace)  printf("%02x,CCR", (unsigned char) temp);

    return SUCCESS;
}


int EORI_SR(void)
{
    long  temp;

    if (!(SR & sbit))
        return NO_PRIVILEGE;

    memRequest(&PC, WORD, &temp);
    SR ^= temp;

    if (trace)  printf("%04x,SR", (unsigned short) temp);

    return SUCCESS_PRIV;
}


int NOT(void)
{
    long  size;

    if (decodeSize(&size))  return BAD_INST;
    if (effAddr (size, DATA_ALT_ADDR))  return BAD_INST;

    valueOf(EA1, &dest, size);
    source = dest;
    result = ~dest;
    put(EA1, result, size);
    ccrUpdate(N_A, GEN, GEN, ZER, ZER, source, dest, result, size, 0);

    return SUCCESS;
}


int SHIFTROT(void)
{
    long  size;
    int   reg, shift_count, shift_size, type, cnt, mem_reg;
    int   direction, temp_bit, temp_bit_2;
    long  ev1;

    if ((mem_reg = ((inst & 0xc0)) == 0xc0)) {
        if (effAddr(WORD, MEM_ALT_ADDR))  return BAD_INST;
        size = WORD;
        shift_count = 1;
        valueOf(EA1, &ev1, size);
        source = dest = ev1;
        type = (inst & 0x0600) >> 9;
    }
    else  {
        if (decodeSize(&size))  return BAD_INST;
        if (inst & 0x0020)
            shift_count = D[(inst >> 9) & 0x7] % 64;
        else {
            shift_count = (inst >> 9) & 0x7;
            if (shift_count == 0)
                shift_count = 8;
        }
        reg = inst & 0x0007;
        source = dest = D[reg] & size;
        type = (inst & 0x18) >> 3;
        EA1 = &D[reg];
        valueOf(EA1, &ev1, size);
    }
    direction = inst & 0x0100;
    if (size == LONG)
        shift_size = 31;
    else if (size == WORD)
        shift_size = 15;
    else
        shift_size = 7;

    if (shift_count == 0)  {
        if (type == 2)
            ccrUpdate(N_A, GEN, GEN, ZER, CASE_1, source, dest, ev1, size, shift_count);
        else
            ccrUpdate(N_A, GEN, GEN, ZER, ZER, source, dest, ev1, size, shift_count);
    }
    else  {
        switch (type) {
            case 0 :
              /* Arithmetic shift */
              if (direction) {
                  /* Shift left */
                  put(EA1, (ev1 & size) << shift_count, size);
                  valueOf (EA1, &ev1, size);
                  ccrUpdate(GEN, GEN, GEN, CASE_4, CASE_3, source, dest, ev1, size, shift_count);
                  if (trace)  printf("asl  ");
              }
              else {
                  /* Shift right */
                  if ((ev1 >> shift_size) & 1)
                      temp_bit = 1;
                  else
                      temp_bit = 0;
                  for (cnt = 1; cnt <= shift_count; cnt++)  {
                      put(EA1, (ev1 & size) >> 1, size);
                      valueOf(EA1, &ev1, size);
                      if (temp_bit)
                          put(EA1, ev1 | (1 << shift_size), size);
                      else
                          put(EA1, ev1 & (~(1 << shift_size)), size);
                      valueOf(EA1, &ev1, size);
                  }
                  ccrUpdate (GEN, GEN, GEN, ZER, CASE_2, source, dest, ev1, size, shift_count);
                  if (trace)  printf("asr  ");
              }
              break;

            case 1 :
              /* Logical shift */
              if (direction)  {
                  /* Shift left */
                  put(EA1, (unsigned long)ev1 << shift_count, size);
                  valueOf(EA1, &ev1, size);
                  ccrUpdate(GEN, GEN, GEN, ZER, CASE_3, source, dest, ev1, size, shift_count);
                  if (trace)  printf("lsl  ");
              }
              else {
                  /* Shift right */
                  put(EA1, ((unsigned long)ev1 & size) >> shift_count, size);
                  valueOf(EA1, &ev1, size);
                  ccrUpdate (GEN, GEN, GEN, ZER, CASE_2, source, dest, ev1, size, shift_count);
                  if (trace)  printf("lsr  ");
              }
              break;

            case 2 :
              /* Rotate with extend */
              if (direction) {
                  /* Rotate left */
                  for (cnt = 1; cnt <= shift_count; cnt++)  {
                      temp_bit = (ev1 >> shift_size) & 1;
                      temp_bit_2 = (SR & xbit) >> 4;
                      put(EA1, (ev1 & size) << 1, size);
                      valueOf(EA1, &ev1, size);
                      if (temp_bit_2)
                          put (EA1, ev1 | 1, size);
                      else
                          put (EA1, ev1 & ~1, size);
                      valueOf(EA1, &ev1, size);
                      if (temp_bit)
                          SR = SR | xbit;
                      else
                          SR = SR & ~xbit;
                  }
                  ccrUpdate(GEN, GEN, GEN, ZER, CASE_3, source, dest, ev1, size, shift_count);
                  if (trace)  printf("roxl ");
              }
              else {
                  /* Rotate right */
                  for (cnt = 1; cnt <= shift_count; cnt++)  {
                      temp_bit = ev1 & 1;
                      temp_bit_2 = (SR & xbit) >> 4;
                      put(EA1, (ev1 & size) >> 1, size);
                      valueOf (EA1, &ev1, size);
                      if (temp_bit_2)
                          put (EA1, ev1 | (1 << shift_size), size);
                      else
                          put (EA1, ev1 & (~(1 << shift_size)), size);
                      valueOf(EA1, &ev1, size);
                      if (temp_bit)
                          SR = SR | xbit;
                      else
                          SR = SR & ~xbit;
                  }
                  put(EA1, ev1, size);		
                  ccrUpdate(GEN, GEN, GEN, ZER, CASE_2, source, dest, ev1, size, shift_count);
                  if (trace)  printf("roxr ");
              }
              break;

            case 3 :
                /* Rotate */
                if (direction) {
                    /* Rotate left */
                    for (cnt = 1; cnt <= shift_count; cnt++)  {
                        temp_bit = (ev1 >> shift_size) & 1;
                        put(EA1, (ev1 & size) << 1, size);
                        valueOf(EA1, &ev1, size);
                        if (temp_bit)
                            put(EA1, ev1 | 1, size);
                        else
                            put(EA1, ev1 & ~1, size);
                        valueOf(EA1, &ev1, size);
                    }
                    ccrUpdate(N_A, GEN, GEN, ZER, CASE_3, source, dest, ev1, size, shift_count);
                    if (trace)  printf("rol ");
                }
                else {
                    /* Rotate right */
                    for (cnt = 1; cnt <= shift_count; cnt++)  {
                        temp_bit = ev1 & 1;
                        put(EA1, (ev1 & size) >> 1, size);
                        valueOf(EA1, &ev1, size);
                        if (temp_bit)
                            put(EA1, ev1 | (1 << shift_size), size);
                        else
                            put(EA1, ev1 & (~(1 << shift_size)), size);
                        valueOf(EA1, &ev1, size);
                    }
                    ccrUpdate(N_A, GEN, GEN, ZER, CASE_2, source, dest, ev1, size, shift_count);
                    if (trace)  printf("ror ");
                }
                break;
            }
    }

    return SUCCESS;
}


int SWAP(void)
{
    int  reg;

    reg = inst & 0x07;

    D[reg] = ((D[reg] & WORD) * 0x10000) | ((D[reg] & 0xffff0000) / 0x10000);

    ccrUpdate(N_A, GEN, GEN, ZER, ZER, source, dest, D[reg], LONG, 0);

    if (trace)  printf("D%d", reg);

    return SUCCESS;
}


int BITOP(void)
{
	long  size, bit;

	if (trace) { 
	    switch ((inst >> 6) & 0x3) {
		case 0: printf("tst    "); break;
		case 1: printf("chg    "); break;
		case 2: printf("clr    "); break;
		case 3: printf("set    "); break;
	    }
	}

	if (inst & 0x0100) {
		bit = D[(inst >> 9) & 0x07];
		if (trace)  printf("D%ld,", (inst >> 9) & 0x07);
	}
	else {
		memRequest(&PC, WORD, &bit);
		bit = bit & 0xff;
		if (trace)  printf("#%ld,", bit);
	}

	if (effAddr(BYTE, DATA_ADDR))
		return BAD_INST;

	if (inst & 0x0038)  {
		bit &= 0x0007;
		size = BYTE;
	}
	else {
		bit &= 0x001F;
		size = LONG;
	}

	valueOf(EA1, &dest, size);
	if (dest & (1 << bit))
		SR &= ~zbit;
	else
		SR |= zbit;

	switch ((inst >> 6) & 0x3) {
	    case 0 : 			/* perform a bit test operation */
		break;

	    case 1 : 			/* perform a bit change operation */
		if ((dest >> bit) & 1)
			put(EA1, dest & (~(1 << bit)), size);
		else
			put(EA1, dest | (1 << bit), size);
		break;

	    case 2 :			/* perform a bit clear operation */
		put(EA1, dest & (~(1 << bit)), size);
		break;

	    case 3 :			/* perform a bit set operation */
		put(EA1, dest | (1 << bit), size);
		break;
	}

	return SUCCESS;
}


int TAS(void)
{
    if (effAddr(BYTE, DATA_ALT_ADDR))
        return BAD_INST;

    valueOf(EA1, &result, BYTE);
    put(EA1, result | 0x80, BYTE);
    ccrUpdate(N_A, GEN, GEN, ZER, ZER, source, dest, result, BYTE, 0);

    return SUCCESS;
}


int BCC(void)
{
    long  disp;
    int   branch;

    branch = checkCondition((inst & 0x0f00) >> 8);

    if ((disp = (inst & 0x00ff)) == 0)  {
        memGet(PC, WORD, &disp);
        from_2s_comp(disp, WORD, &disp);
        if (branch)
            PC += disp;
        else
            PC += 2;
        if (trace)  printf(".w   %04x", (unsigned short) disp);
    }
    else  {
        from_2s_comp(disp, BYTE, &disp);
        if (branch)
            PC += disp;		
        if (trace)  printf(".s   %02x", (unsigned short) disp);
    }

    return SUCCESS;
}


int DBCC(void)
{
     long  disp;
     int   reg;
     int   oprt;

     memRequest(&PC, WORD, &disp);
     from_2s_comp(disp, WORD, &disp);
     reg = inst & 0x0007;

     if (!checkCondition((inst >> 8) & 0x0f))  {
         oprt = (D[reg] & WORD);
         oprt--;
         put(&D[reg], oprt, WORD);
         if (oprt != -1) {
             PC += (disp - 2);
         }
    }
    if (trace)  printf("   D%d,%04x", reg, (unsigned short) disp);

    return SUCCESS;
}


int SCC(void)
{
    if (effAddr(BYTE, DATA_ALT_ADDR))
        return BAD_INST;

    if (checkCondition((inst & 0x0f00) >> 8))
        put(EA1, 0xff, BYTE);
    else
        put(EA1, 0x00, BYTE);

    return SUCCESS;
}


int BRA(void)
{
    long  disp;

    if ((disp = (inst & 0x00ff)) == 0)  {
        memGet(PC, WORD, &disp);
        from_2s_comp(disp, WORD, &disp);
        PC += disp;
        if (trace)  printf(".w   %04x", (unsigned short) disp);
    }
    else  {
        from_2s_comp(disp, BYTE, &disp);
        PC += disp;
        if (trace)  printf(".s   %02x", (unsigned char) disp);
    }

    return SUCCESS;
}


int BSR(void)
{
    long  disp;

    if ((disp = (inst & 0x00ff)) == 0)  {
        memGet(PC, WORD, &disp);
        from_2s_comp(disp, WORD, &disp);
        A[a_reg(7)] -= 4;
        memPut(PC+2, A[a_reg(7)], LONG);
        PC += disp;
        if (trace)  printf(".w   %04x", (unsigned short) disp);
    }
    else  {
        from_2s_comp(disp, BYTE, &disp);
        A[a_reg(7)] -= 4;
        memPut(PC, A[a_reg(7)], LONG);
        PC += disp;
        if (trace)  printf(".s   %02x", (unsigned char) disp);
    }

    return SUCCESS;
}


int JMP(void)
{
    if (effAddr(WORD, CONTROL_ADDR))
        return BAD_INST;

    PC = *EA1;

    return SUCCESS;
}


int JSR(void)
{
    if (effAddr(WORD, CONTROL_ADDR))
        return BAD_INST;

    A[a_reg(7)] -= 4; memPut(PC, A[a_reg(7)], LONG);
    PC = *EA1;

    return SUCCESS;
}


int RTE(void)
{
    long  temp;

    if (!(SR & sbit))
        return (NO_PRIVILEGE);

    memGet(A[8], WORD, &temp);
    A[8] += 2;
    SR = temp & 0x0000ffff;
    memGet(A[8], LONG, &PC);
    A[8] += 4;

    return SUCCESS_PRIV;
}


int RTR(void)
{
    long  temp;

    memGet(A[a_reg(7)], WORD, &temp);
    A[a_reg(7)] += 2;
    SR = (SR & 0xff00) | (temp & 0x00ff);
    memGet(A[a_reg(7)], LONG, &PC);
    A[a_reg(7)] += 4;

    return SUCCESS_PRIV;
}


int RTS(void)
{
    memGet(A[a_reg(7)], LONG, &PC);
    A[a_reg(7)] += 4;

    return SUCCESS;
}


int NOP(void)
{
    return SUCCESS;
}


int CHK(void)
{
    int   reg;

    reg = (inst >> 9) & 0x07;

    if (effAddr(WORD, DATA_ADDR))
        return BAD_INST;

    valueOf(EA1, &source, WORD);
    from_2s_comp(source, WORD, &source);
    dest = D[reg] & WORD;

    ccrUpdate(N_A, GEN, UND, UND, UND, source, D[reg], D[reg], WORD, 0);

    if ((dest < 0) || (dest > source))
        return CHK_EXCEPTION;

    return SUCCESS;
}


int ILLEGAL(void)
{
    return ILLEGAL_TRAP;
}


int RESET(void)
{
    if (!(SR & sbit))
        return NO_PRIVILEGE;

    return SUCCESS;
}



int STOP(void)
{
    long  temp;
    int   tr_on;

    memRequest(&PC, WORD, &temp);

    tr_on = (SR & tbit) ? 1 : 0;

    if (!(SR & sbit))
        return NO_PRIVILEGE;

    SR = temp & WORD;
    if (tr_on)
        SR = SR | tbit;

    if (!(SR & sbit))
        return NO_PRIVILEGE;

    cpuState = ST_STOP;

    return SUCCESS_PRIV;
}



int TRAP(void)
{
    if (trace)  printf("#%d\n", (int)(inst & 0x000F));

    return TRAP_TRAP;
}



int TRAPV(void)
{
    if (SR & vbit)
        return TRAPV_TRAP;

     return SUCCESS;
}


/*
 * Instruction table
 */
Inst instTable[] = {
    { 0x0000, 0x0a3c, EORI_CCR, "eori.b  " },    /* 0 */
    { 0x0000, 0x0a7c, EORI_SR,  "eori.w  " },
    { 0x0000, 0x003c, ORI_CCR,  "ori.b   " },
    { 0x0000, 0x007c, ORI_SR,   "ori.w   " },
    { 0x0000, 0x023c, ANDI_CCR, "andi.b  " },
    { 0x0000, 0x027c, ANDI_SR,  "andi.w  " },
    { 0x0fc7, 0x0008, MOVEP,    "movep   " },
    { 0x00ff, 0x0200, ANDI,     "andi    " },
    { 0x00ff, 0x0000, ORI,      "ori     " },
    { 0x00ff, 0x0400, SUBI,     "subi    " },
    { 0x00ff, 0x0600, ADDI,     "addi    " },
    { 0x00ff, 0x0800, BITOP,    "b"        },
    { 0x00ff, 0x0a00, EORI,     "eori    " },
    { 0x00ff, 0x0c00, CMPI,     "cmpi    " },
    { 0x0eff, 0x0100, BITOP,    "b"        },

    { 0x0fff, 0x1000, MOVE,     "move.b  " },   /* 15 */

    { 0x0e3f, 0x2040, MOVEA,    "movea.l " },   /* 16 */
    { 0x0fff, 0x2000, MOVE,     "move.l  " },

    { 0x0e3f, 0x3040, MOVEA,    "movea.l " },   /* 18 */
    { 0x0fff, 0x3000, MOVE,     "move.w  " },

    { 0x0000, 0x4afc, ILLEGAL,  "illegal " },   /* 20 */
    { 0x0000, 0x4e70, RESET,    "reset   " },
    { 0x0000, 0x4e71, NOP,      "nop     " },
    { 0x0000, 0x4e72, STOP,     "stop    " },
    { 0x0000, 0x4e73, RTE,      "rte     " },
    { 0x0000, 0x4e75, RTS,      "rts     " },
    { 0x0000, 0x4e76, TRAPV,    "trapv   " },
    { 0x0000, 0x4e77, RTR,      "rtr     " },
    { 0x0007, 0x4e50, LINK,     "link    " },
    { 0x0007, 0x4e58, UNLK,     "unlk    " },
    { 0x0007, 0x4e60, MOVE_USP, "move.l  " },
    { 0x0007, 0x4e68, MOVE_USP, "move.l  " },
    { 0x0007, 0x4840, SWAP,     "swap    " },
    { 0x0007, 0x4880, EXT,      "ext     " },
    { 0x0007, 0x48c0, EXT,      "ext     " },
    { 0x000f, 0x4e40, TRAP,     "trap    " },
    { 0x003f, 0x40c0, MOVE_SR_, "move.w  " },
    { 0x003f, 0x44c0, MOVE_CCR, "move.b  " },
    { 0x003f, 0x46c0, MOVE_SR,  "move.w  " },
    { 0x003f, 0x4800, NBCD,     "nbcd    " },
    { 0x003f, 0x4840, PEA,      "pea     " },
    { 0x003f, 0x4ac0, TAS,      "tas     " },
    { 0x003f, 0x4e80, JSR,      "jsr     " },
    { 0x003f, 0x4ec0, JMP,      "jmp     " },
    { 0x007f, 0x4c80, MOVEM,    "movem   " },
    { 0x007f, 0x4880, MOVEM,    "movem   " },
    { 0x00ff, 0x4200, CLR,      "clr     " },
    { 0x00ff, 0x4000, NEGX,     "negx    " },
    { 0x00ff, 0x4400, NEG,      "neg     " },
    { 0x00ff, 0x4600, NOT,      "not     " },
    { 0x00ff, 0x4a00, TST,      "tst     " },
    { 0x0e3f, 0x4180, CHK,      "chk     " },
    { 0x0e3f, 0x41c0, LEA,      "lea     " },

    { 0x0f07, 0x50c8, DBCC,     "db"       },   /* 53 */
    { 0x0f3f, 0x50c0, SCC,      "s"        },
    { 0x0eff, 0x5100, SUBQ,     "subq    " },
    { 0x0eff, 0x5000, ADDQ,     "addq    " },

    { 0x00ff, 0x6000, BRA,      "bra"      },   /* 57 */
    { 0x00ff, 0x6100, BSR,      "bsr"      },
    { 0x0fff, 0x6000, BCC,      "b"        },

    { 0x0eff, 0x7000, MOVEQ,    "moveq   " },   /* 60 */

    { 0x0e0f, 0x8100, SBCD,     "sbcd    " },   /* 61 */
    { 0x0e3f, 0x80c0, DIVU,     "divu    " },
    { 0x0e3f, 0x81c0, DIVS,     "divs    " },
    { 0x0fff, 0x8000, OR,       "or      " },

    { 0x0f3f, 0x90c0, SUBA,     "suba    " },   /* 65 */
    { 0x0ecf, 0x9100, SUBX,     "subx    " },
    { 0x0fff, 0x9000, SUB,      "sub     " },

    { 0x0f3f, 0xb0c0, CMPA,     "cmpa    " },   /* 68 */
    { 0x0ec7, 0xb108, CMPM,     "cmpm    " },
    { 0x0eff, 0xb000, CMP,      "cmp     " },
    { 0x0eff, 0xb100, EOR,      "eor     " },

    { 0x0e07, 0xc140, EXG,      "exg     " },   /* 72 */
    { 0x0e07, 0xc148, EXG,      "exg     " },
    { 0x0e07, 0xc188, EXG,      "exg     " },
    { 0x0e0f, 0xc100, ABCD,     "abcd    " },
    { 0x0e3f, 0xc1c0, MULS,     "muls    " },
    { 0x0e3f, 0xc0c0, MULU,     "mulu    " },
    { 0x0fff, 0xc000, AND,      "and     " },

    { 0x0f3f, 0xd0c0, ADDA,     "adda    " },   /* 79 */
    { 0x0ecf, 0xd100, ADDX,     "addx    " },
    { 0x0fff, 0xd000, ADD,      "add     " },

    { 0x073f, 0xe0c0, SHIFTROT, ""         },   /* 82 */
    { 0x0fff, 0xe000, SHIFTROT, ""         },
};

/*
 * Offsets into the instruction table
 *
 * Each offset marks the begin of the opcode with the
 * corresponding first digit.
 */
int  instOffset[] = {
    0, 15, 16, 18, 20, 53, 57, 60, 61, 65, 68, 68, 72, 79, 82, 84, 83
};

